
from .models.Address import Address
from .models.OrderUnit import OrderUnit
from .models.InvoiceItem import InvoiceItem
from .models.Location import Location
from .models.LineItem import LineItem
from .models.CustomFieldsToCreate import CustomFieldsToCreate
from .refresher import refreshCustomFormsAndFields
from .outputters.csvWriter import writeListOfObjectsToCsvWithObjectPropertiesAsColumnNames
from .outputters.emailer import sendEmailUsingGmailCredentialsWithFilesAttached


class OrderUnit:
	def __init__(self,name:str,multiplier:int=1,active:bool=True):
		self.id=None
		self.name=name
		self.multiplier=multiplier
		self.active=active
	def getJsonReadyValue(self):
		return self.__dict__


from .WriteOperationCls import WriteOperation
from .writeHelpers import getWriteOperationToUpdateObject,getBodyForWriteOperation,addWarehouseAndLabelToStockLevelUpdateIfPresent
from .execute import executeWrites
from typing import Union,Literal,Union
from ..models.Invoice import Invoice
from ..models.InvoiceItem import InvoiceItem
from datetime import datetime
from typing import List
from ..models.CustomFieldsToCreate import CustomFieldsToCreate
from ..models.Warehouse import Warehouse


class WriterBase:
	def __init__(self):
		self.writeOperations:List[WriteOperation]=[]
	def apply(self):
		res= executeWrites(self.writeOperations)
		self.writeOperations=[]
		return res
	##unchanging update operations
	def addStockLevelUpdate(self,product_id:int,order_unit_id:int,level:int,warehouse_id:int=0,label:Union[None,str]=None):
		item={'product_id': product_id, 'order_unit_id': order_unit_id, 'level': level}
		addWarehouseAndLabelToStockLevelUpdateIfPresent(item,warehouse_id,label)
		self.writeOperations.append(WriteOperation("stocklevels", "post", item))

	def addStockLevelUpdateUsingProductCodeAndUnitName(self,product_code:str,order_unit_name:str,level:int,warehouse_id:int=0,label:Union[None,str]=None):
		item={'product_code': product_code, 'order_unit_name': order_unit_name, 'level': level}
		addWarehouseAndLabelToStockLevelUpdateIfPresent(item,warehouse_id,label)
		self.writeOperations.append(WriteOperation("stocklevels", "post", item))

	def addPriceUpdate(self,product_id:int,order_unit_id:int,price:float,price_list_id:int,tax_rate_id:Union[None,int]=None):
		item= {'product_id': product_id, 'order_unit_id': order_unit_id, 'price': price, 'price_list_id': price_list_id}
		if tax_rate_id!=None:
			item['tax_rate_id']=tax_rate_id
		self.writeOperations.append(WriteOperation("prices", "post",item))

	def addPriceUpdateUsingProductCodeAndUnitName(self,product_code:str,order_unit_name:str,price:float,price_list_id:int,tax_rate_id:Union[None,int]=None):
		item= {'product_code': product_code, 'order_unit_name': order_unit_name, 'price': price, 'price_list_id': price_list_id}
		if tax_rate_id!=None:
			item['tax_rate_id']=tax_rate_id
		self.writeOperations.append(WriteOperation("prices", "post",item))

	def addInvoiceUpdate(self,invoice:Invoice,fieldsToUpdate:List[str]):
		self.writeOperations.append(getWriteOperationToUpdateObject(invoice,fieldsToUpdate))
	##unchanging create operations
	def addInvoiceCreate(self,date:datetime,customer_code:str,items:List[InvoiceItem],reference='',status:Union[None,Literal['Draft','Authorized','Delivered','Outstanding','Paid','Deleted']]=None,due_date:Union[None,datetime]=None,taxIsIncludedInLineValues=True,outstanding_balance:Union[None,float]=None):
		body=getBodyForWriteOperation(locals())
		del body['taxIsIncludedInLineValues']
		if not(taxIsIncludedInLineValues):
			for item in body['items']:
				taxAmount=0
				if 'tax_amount' in item:
					taxAmount=item['tax_amount']
				item['value']=item['value']+taxAmount
		self.writeOperations.append(WriteOperation("invoices", "post", body))

	def addScheduledVisitCreate(self,assigned_user_name:str,customer_code:str,due_date:datetime,end_time:Union[datetime,None]=None,comment:Union[None,str]=None):
		body=getBodyForWriteOperation(locals())
		if end_time==None:
			body['all_day']=True
		self.writeOperations.append(WriteOperation("scheduledvisits", "post", body))

	def addCustomFieldCreations(self,customFieldsToCreate:CustomFieldsToCreate):
		self.writeOperations.append(WriteOperation("integrations", "post", {'action':'AddCustomFields','fields_to_add':customFieldsToCreate.fields_to_add},canBeCombinedWithOtherWritesInAList=False))

	def addWarehouseCreate(self,name:str,order_email:str='',credit_request_email:str='',quote_email:str='',active:bool=True):
		body=getBodyForWriteOperation(locals())
		self.writeOperations.append(WriteOperation("warehouses", "post", body))

	def addWarehouseUpdate(self,warehouse:Warehouse,fieldsToUpdate:List[Literal['name','order_email','credit_request_email','quote_email','active']]):
		body={'id':warehouse.id,'name':warehouse.name}
		for field in fieldsToUpdate:
			body[field]=getattr(warehouse,field)
		self.writeOperations.append(WriteOperation("warehouses", "patch", body))


from ..shared.helpers import getDateTimeObjectFromSkynamoDateTimeStr
from datetime import datetime

class Warehouse:
	def __init__(self,json:dict):
		self.id:int=json['id']
		self.name:str=json['name']
		self.order_email:str=json['order_email']
		self.credit_request_email:str=json['credit_request_email']
		self.quote_email:str=json['quote_email']
		self.active:bool=json['active']
		self.last_modified_time:datetime=getDateTimeObjectFromSkynamoDateTimeStr(json['last_modified_time'])


from .CreditRequest import CreditRequest
from .Order import Order
from .Quote import Quote
from .Product import Product
from .Customer import Customer
from skynamo.models.Invoice import Invoice
from skynamo.models.User import User
from skynamo.models.Warehouse import Warehouse
from skynamo.models.Price import Price
from skynamo.models.StockLevel import StockLevel
from skynamo.reader.jsonToObjects import getListOfObjectsFromJsonFile,populateCustomPropsFromFormResults,populateUserIdAndNameFromInteractionAndReturnFormIds
from skynamo.reader.sync import refreshJsonFilesLocallyIfOutdated,getSynchedDataTypeFileLocation
import json
from typing import List,Dict
##|customImports|##

def _getTransactions(transactionClass,forceRefresh=False):
	refreshJsonFilesLocallyIfOutdated([f'{transactionClass.__name__.lower()}s','completedforms','interactions'])#type:ignore
	interactionsJson={}
	with open(getSynchedDataTypeFileLocation('interactions'), "r") as read_file:
		interactionsJson=json.load(read_file)
	refreshJsonFilesLocallyIfOutdated(['orders','completedforms','interactions'],forceRefresh)
	completedForms={}
	with open(getSynchedDataTypeFileLocation('completedforms'), "r") as read_file:
		completedForms=json.load(read_file)
	transactions=getListOfObjectsFromJsonFile(getSynchedDataTypeFileLocation(f'{transactionClass.__name__.lower()}s'),transactionClass)
	for i,transaction in enumerate(transactions):
		formIds=populateUserIdAndNameFromInteractionAndReturnFormIds(transaction,interactionsJson)
		populateCustomPropsFromFormResults(transaction,formIds,completedForms)
	return transactions

class Reader:
	def __init__(self):
		pass
	def getOrders(self,forceRefresh=False):
		orders:List[Order]=_getTransactions(Order,forceRefresh)
		return orders

	def getCreditRequests(self,forceRefresh=False):
		creditRequests:List[CreditRequest]=_getTransactions(CreditRequest,forceRefresh)
		return creditRequests

	def getQuotes(self,forceRefresh=False):
		quotes:List[Quote]=_getTransactions(Quote,forceRefresh)
		return quotes

	def getProducts(self,forceRefresh=False):
		refreshJsonFilesLocallyIfOutdated(['products'],forceRefresh)
		products:List[Product]= getListOfObjectsFromJsonFile(getSynchedDataTypeFileLocation('products'),Product)
		return products

	def getCustomers(self,forceRefresh=False):
		refreshJsonFilesLocallyIfOutdated(['customers'],forceRefresh)
		customers:List[Customer]= getListOfObjectsFromJsonFile(getSynchedDataTypeFileLocation('customers'),Customer)
		return customers

	def getInvoices(self,forceRefresh=False):
		refreshJsonFilesLocallyIfOutdated(['invoices'],forceRefresh)
		invoices:List[Invoice]=getListOfObjectsFromJsonFile(getSynchedDataTypeFileLocation('invoices'),Invoice)
		return invoices

	def getStockLevels(self,forceRefresh=False):
		refreshJsonFilesLocallyIfOutdated(['stocklevels'],forceRefresh)
		stockLevels:List[StockLevel]= getListOfObjectsFromJsonFile(getSynchedDataTypeFileLocation('stocklevels'),StockLevel)
		return stockLevels

	def getUsers(self,forceRefresh=False):
		refreshJsonFilesLocallyIfOutdated(['users'],forceRefresh)
		users:List[User]= getListOfObjectsFromJsonFile(getSynchedDataTypeFileLocation('users'),User)
		return users

	def getWarehouses(self,forceRefresh=False):
		refreshJsonFilesLocallyIfOutdated(['warehouses'],forceRefresh)
		warehouses:List[Warehouse]= getListOfObjectsFromJsonFile(getSynchedDataTypeFileLocation('warehouses'),Warehouse)
		return warehouses

	def getPrices(self,forceRefresh=False):
		refreshJsonFilesLocallyIfOutdated(['prices'],forceRefresh)
		prices:List[Price]= getListOfObjectsFromJsonFile(getSynchedDataTypeFileLocation('prices'),Price)
		return prices





class User:
	def __init__(self,json:dict):
		self.id:int=json['id']
		self.user_name:str=json['user_name']
		self.display_name:str=json['display_name']
		self.email:str=json['email']
		self.is_verified:bool=json['is_verified']
		self.active:bool=json['active']
		if 'access' in json:
			self.access:str=json['access']



from typing import Literal,Union,List
from datetime import datetime
from ..shared.helpers import getDateTimeObjectFromSkynamoDateTimeStr
from .InvoiceItem import InvoiceItem

class Invoice:
	def __init__(self,json:dict={}):
		self.id:int=json['id']
		self.date:datetime=getDateTimeObjectFromSkynamoDateTimeStr(json['date'])
		self.customer_id:int=json['customer_id']
		self.customer_code:str=json['customer_code']
		self.reference:Union[str,None]=None
		if 'reference' in json:
			self.reference=json['reference']
		self.row_version:int=json['row_version']
		self.last_modified_time:datetime=getDateTimeObjectFromSkynamoDateTimeStr(json['last_modified_time'])
		self.status:Union[None,Literal['Draft','Authorized','Delivered','Outstanding','Paid','Deleted']]=None
		if 'status' in json:
			self.status=json['status']
		self.due_date:Union[None,datetime]=None
		if 'due_date' in json:
			self.due_date=getDateTimeObjectFromSkynamoDateTimeStr(json['due_date'])
		self.external_id:Union[None,str]=None
		if 'external_id' in json:
			self.external_id=json['external_id']
		self.tax_inclusion:Union[None,Literal['Included','Excluded']]=None
		if 'tax_inclusion' in json:
			self.tax_inclusion=json['tax_inclusion']
		self.total_tax_amount:Union[None,float]=None
		if 'tax' in json:
			self.total_tax_amount=json['tax']
		self.outstanding_balance:Union[None,float]=None
		if 'outstanding_balance' in json:
			self.outstanding_balance=json['outstanding_balance']
		self.items:List[InvoiceItem]=[]
		for item in json['items']:
			inputToInvoiceItem={'product_id':item['product_id'],'product_code':item['product_code'],'quantity':item['quantity'],'totalLineValue':item['value']}
			if 'tax_amount' in item:
				inputToInvoiceItem['tax_amount']=item['tax_amount']
			self.items.append(InvoiceItem(**inputToInvoiceItem))


from .WriteOperationCls import WriteOperation
from .WriteErrorCls import WriteError
from ..shared.api import makeRequest
import math,threading
from typing import List,Union

def executeWrites(writeOperations:List[WriteOperation]):
	writeBatchesGroupedByDataTypeAndHttpMethod=[]
	for write in writeOperations:
		found=False
		if not(write.canBeCombinedWithOtherWritesInAList):
			writeBatchesGroupedByDataTypeAndHttpMethod.append(write)
			continue
		for writeBatch in writeBatchesGroupedByDataTypeAndHttpMethod:
			if writeBatch[0].dataType==write.dataType and writeBatch[0].httpMethod==write.httpMethod:
				writeBatch.append(write)
				found=True
				break
		if not found:
			writeBatchesGroupedByDataTypeAndHttpMethod.append([write])
	subBatchesWithMaxSizeOf20:List[List[WriteOperation]]=[]
	for writeBatch in writeBatchesGroupedByDataTypeAndHttpMethod:
		if not(isinstance(writeBatch,list)):
			subBatchesWithMaxSizeOf20.append(writeBatch)
			continue
		for i in range(math.ceil(len(writeBatch)/20)):
			subBatchesWithMaxSizeOf20.append(writeBatch[i*20:i*20+20])
	return __makeThreadedWrites(subBatchesWithMaxSizeOf20)

def __makeThreadedWrites(subBatchesWithMaxSizeOf20:List[List[WriteOperation]]):
	threads=[]
	errors:List[WriteError]=[]
	for subBatch in subBatchesWithMaxSizeOf20:
		threads.append(threading.Thread(target=__makeWriteRequest,args=(subBatch,errors)))
	for thread in threads:
		thread.start()
	for thread in threads:
		thread.join()
	return errors

def __makeWriteRequest(writeOperations:Union[WriteOperation,List[WriteOperation]],errors:List[WriteError]):
	body=[]
	httpMethod=''
	dataType=''
	if isinstance(writeOperations,list):
		httpMethod=writeOperations[0].httpMethod
		dataType=writeOperations[0].dataType
		for write in writeOperations:
			body.append(write.itemOrId)
	else:
		body=writeOperations.itemOrId
		httpMethod=writeOperations.httpMethod
		dataType=writeOperations.dataType
	results=makeRequest(httpMethod,dataType,body)#type:ignore
	if 'errors' in results:
		for error in results['errors']:
			errors.append(WriteError(dataType,httpMethod,body,error['detail']))


from typing import Union

class InvoiceItem:
	def __init__(self,product_code:str,quantity:float,totalLineValue:float,tax_amount:Union[None,float]=None,product_id:Union[None,int]=None):
		self.product_id=product_id
		self.product_code:str=product_code
		self.quantity:float=quantity
		self.totalLineValue:float=totalLineValue
		self.tax_amount:Union[None,float]=tax_amount

	def getJsonReadyValue(self):
		res={
			'product_id':self.product_id,
			'product_code':self.product_code,
			'quantity':self.quantity,
			'value':self.totalLineValue
		}
		if self.tax_amount != None:
			res['tax_amount']=self.tax_amount
		return res


from setuptools import setup, find_packages
import os
lib_folder = os.path.dirname(os.path.realpath(__file__))
requirement_path = lib_folder + '/requirements.txt'
install_requires = []
if os.path.isfile(requirement_path):
    with open(requirement_path) as f:
        for dependancy in f.read().splitlines():
            if dependancy[:4]=='git+':
                packageName=dependancy.split('/')[-1]
                dependancy=packageName+" @ "+dependancy
            install_requires.append(dependancy)

if __name__ == "__main__":
    setup(
        name='skynamo',
        version='0.0.9',
        author='Daniel van Niekerk',
        author_email='daniel@skynamo.com',
        description='Skynamo Public API SDK',
        url='https://github.com/skynamo/skynamo-python-sdk',
        packages=find_packages(),
        install_requires=install_requires
    )